"""
Standard Regression model
-------------------------
"""

from .regression_model import RegressionModel
from ..timeseries import TimeSeries
from ..logging import get_logger, raise_log, raise_if
from typing import List, Optional
from warnings import warn
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

logger = get_logger(__name__)


class StandardRegressionModel(RegressionModel):

    def __init__(self,
                 train_n_points: Optional[int] = None,
                 model=LinearRegression(n_jobs=-1, fit_intercept=False)):
        """
        Simple wrapper for regression models implementing a fit() predict() functions
        (e.g., scikit-learn regression).

        Parameters
        ----------
        train_n_points
            The number of most recent points from the training (features and target) time series that
            will be used to train the regression model. If this is `None`, or if the provided training series
            contain fewer points, the maximum possible number of time steps will be used for training.
        model
            A regression model that implements `fit()` and `predict()` methods.
            Default: `sklearn.linear_model.LinearRegression(n_jobs=-1, fit_intercept=False)`
        """

        super().__init__()
        if (not callable(getattr(model, "fit", None))):
            raise_log(Exception('Provided model object must have a fit() method', logger))
        if (not callable(getattr(model, "predict", None))):
            raise_log(Exception('Provided model object must have a predict() method', logger))

        raise_if(train_n_points is not None and train_n_points < 1,
                 "train_n_points should be greater or equal 1 (or None)",
                 logger)

        self.train_n_points = train_n_points
        self.model = model

    @staticmethod
    def _get_features_matrix_from_series(features: List[TimeSeries]):
        return np.concatenate([s.values() for s in features], axis=1)  # (n_samples x n_features)

    def fit(self,
            train_features: List[TimeSeries],
            train_target: TimeSeries) -> None:

        if self.train_n_points is None:
            train_n_points = min([len(s) for s in train_features] + [len(train_target)])
        else:
            train_n_points = self.train_n_points

            # display a warning when fitting on less than train_n_points training points
            max_train_points = len(train_features[0])
            if train_n_points > max_train_points:
                warn("fitting standard regression model on less than train_n_points training points")

        # Get (at most) the last [train_n_points] of each series
        last_train_ts = train_features[0].end_time()
        last_n_points_features = [s.slice_n_points_before(last_train_ts, train_n_points) for s in train_features]
        last_n_points_target = train_target.slice_n_points_before(last_train_ts, train_n_points)

        super().fit(last_n_points_features, last_n_points_target)

        self.model.fit(self._get_features_matrix_from_series(last_n_points_features),
                       last_n_points_target.values())

    def predict(self, features: List[TimeSeries]) -> TimeSeries:
        super().predict(features)
        y = self.model.predict(self._get_features_matrix_from_series(features))
        return TimeSeries(pd.DataFrame(y, index=features[0].time_index()), self.train_target.freq())



"""
Regression ensemble model
-------------------------
"""
from sklearn.linear_model import LinearRegression
from typing import Optional, List

from darts.timeseries import TimeSeries
from darts.models import EnsembleModel, StandardRegressionModel
from darts.models.forecasting_model import ForecastingModel
from darts.logging import get_logger, raise_if

logger = get_logger(__name__)


class RegressionEnsembleModel(EnsembleModel):
    def __init__(self,
                 forecasting_models: List[ForecastingModel],
                 regression_train_n_points: int,
                 regression_model=LinearRegression(n_jobs=-1, fit_intercept=False)):
        """
        Class for ensemble models using a regression model for ensembling individual models' predictions
        The provided regression model must implement fit() and predict() methods
        (e.g. scikit-learn regression models)

        Parameters
        ----------
        forecasting_models
            List of forecasting models whose predictions to ensemble
        regression_train_n_points
            The number of points to use to train the regression model
        regression_model
            Any regression model with predict() and fit() methods (e.g. from scikit-learn)
            Default: `sklearn.linear_model.LinearRegression(n_jobs=-1, fit_intercept=False)`
        """
        super().__init__(forecasting_models)

        # wrap provided regression_model in a StandardRegressionModel (if not already the case)
        if isinstance(regression_model, StandardRegressionModel):
            # raise exception if train_n_points value is ambiguous
            model_train_n_points = regression_model.train_n_points
            raise_if(model_train_n_points is not None and regression_train_n_points != model_train_n_points,
                     "Provided StandardRegressionModel.train_n_points parameter doesn't match specified"
                     " regression_train_n_points parameter.",
                     logger)

            # if it was None, set regression_model.train_n_points to regression_train_n_points
            regression_model.train_n_points = regression_train_n_points
        else:
            regression_model = StandardRegressionModel(regression_train_n_points, regression_model)

        self.regression_model = regression_model

    def fit(self, series: TimeSeries) -> None:
        super().fit(series)

        # spare train_n_points points to serve as regression target
        raise_if(len(self.training_series) <= self.regression_model.train_n_points,
                 "regression_train_n_points parameter too big (must be smaller or equal" +
                 " to the number of points in training_series)",
                 logger)
        forecast_training = self.training_series[:-self.regression_model.train_n_points]
        regression_target = self.training_series[-self.regression_model.train_n_points:]

        # fit the forecasting models
        for model in self.models:
            model.fit(forecast_training)

        # predict train_n_points points for each model
        predictions = []
        for model in self.models:
            predictions.append(model.predict(self.regression_model.train_n_points))

        # train the regression model on the individual models' predictions
        self.regression_model.fit(train_features=predictions, train_target=regression_target)

        # prepare the forecasting models for further predicting by fitting
        # them with the entire data

        # Some models (incl. Neural-Network based models) may need to be 'reset'
        # to allow being retrained from scratch
        self.models = [model.untrained_model() if hasattr(model, 'untrained_model') else model
                       for model in self.models]

        # fit the forecasting models
        for model in self.models:
            model.fit(self.training_series)

    def ensemble(self, predictions: List[TimeSeries]) -> TimeSeries:
        return self.regression_model.predict(predictions)



"""
Invertible Data Transformer
---------------------------
"""
from abc import abstractmethod
from typing import TypeVar

from darts.logging import get_logger, raise_if_not
from darts.dataprocessing.transformers import BaseDataTransformer

logger = get_logger(__name__)
T = TypeVar('T')


class InvertibleDataTransformer(BaseDataTransformer[T]):

    def __init__(self,
                 name: str = "InvertibleDataTransformer"):

        """
        Abstract class for data transformers implementing a fit method. All deriving classes must implement
        `transform()` and `inverse_transform()`.

        Parameters
        ----------
        names
            The data transformer's name
        """
        super().__init__(name)

    @abstractmethod
    def inverse_transform(self, data: T, *args, **kwargs) -> T:
        """
        Perform inverse transformation of the data. Not implemented in base class.

        Parameters
        ----------
        data
            Object which will be inverse transformed.
        args
            Additional positional arguments for the `inverse_transform()` method
        kwargs
            Additional keyword arguments for the `inverse_transform()` method

        Returns
        -------
        T
            Inverse transformed data.
        """
        if hasattr(self, "_fit_called"):
            raise_if_not(self._fit_called, "fit() must have been called before inverse_transform()", logger)



"""
Mappers
-------
"""
import numpy as np
import pandas as pd

from typing import Callable, Union

from darts.timeseries import TimeSeries
from darts.dataprocessing.transformers import BaseDataTransformer, InvertibleDataTransformer
from darts.logging import get_logger

logger = get_logger(__name__)


class Mapper(BaseDataTransformer[TimeSeries]):
    def __init__(self,
                 fn: Union[Callable[[np.number], np.number], Callable[[pd.Timestamp, np.number], np.number]],
                 name: str = "Mapper"):
        """
        Data transformer to apply a function to a time series (similar to calling `series.map()`)

        Parameters
        ----------
        fn
            Either a function which takes a value and returns a value ie. f(x) = y
            Or a function which takes a value and its timestamp and returns a value ie. f(timestamp, x) = y
        name
            A specific name for the transformer
        """
        super().__init__(name)
        self._fn = fn

    def transform(self, data: TimeSeries, *args, **kwargs) -> TimeSeries:
        super().transform(data)
        return data.map(self._fn)


class InvertibleMapper(InvertibleDataTransformer[TimeSeries]):
    def __init__(self,
                 fn: Union[Callable[[np.number], np.number], Callable[[pd.Timestamp, np.number], np.number]],
                 inverse_fn: Union[Callable[[np.number], np.number], Callable[[pd.Timestamp, np.number], np.number]],
                 name: str = "InvertibleMapper"):
        """
        Data transformer to apply a function and its inverse to a time series (similar to calling `series.map()`)

        Parameters
        ----------
        fn
            Either a function which takes a value and returns a value ie. f(x) = y
            Or a function which takes a value and its timestamp and returns a value ie. f(timestamp, x) = y
        inverse_fn
            Similarly to `fn`, either a function which takes a value and returns a value ie. f(x) = y
            Or a function which takes a value and its timestamp and returns a value ie. f(timestamp, x) = y
            `inverse_fn` should be such that `inverse_fn(fn(x)) == x`
        name
            A specific name for the transformer
        """
        super().__init__(name)
        self._fn = fn
        self._inverse_fn = inverse_fn

    def transform(self, data: TimeSeries, *args, **kwargs) -> TimeSeries:
        super().transform(data)
        return data.map(self._fn)

    def inverse_transform(self, data: TimeSeries, *args, **kwargs):
        super().inverse_transform(data, *args, *kwargs)
        return data.map(self._inverse_fn)



"""
Pipeline
--------
"""
from copy import deepcopy
from typing import Sequence, Union, Iterator

from darts.logging import raise_if_not, get_logger
from darts import TimeSeries
from darts.dataprocessing.transformers import BaseDataTransformer, InvertibleDataTransformer, FittableDataTransformer

logger = get_logger(__name__)


class Pipeline:
    def __init__(self,
                 transformers: Sequence[BaseDataTransformer[TimeSeries]],
                 copy: bool = False):
        """
        Pipeline combines multiple data transformers chaining them together.

        Parameters
        ----------
        transformers
            Sequence of data transformers.
        copy
            If set makes a (deep) copy of each data transformer before adding them to the pipeline
        """
        raise_if_not(all((isinstance(t, BaseDataTransformer)) for t in transformers),
                     "transformers should be objects deriving from BaseDataTransformer", logger)

        if transformers is None or len(transformers) == 0:
            logger.warning("Empty pipeline created")
            self._transformers: Sequence[BaseDataTransformer[TimeSeries]] = []
        elif copy:
            self._transformers = deepcopy(transformers)
        else:
            self._transformers = transformers

        self._invertible = all((isinstance(t, InvertibleDataTransformer) for t in self._transformers))

    def fit(self, data: TimeSeries):
        """
        Fit all fittable transformers in pipeline.

        Parameters
        ----------
        data
            TimeSeries to fit on.
        """

        # Find the last fittable transformer index
        # No need to fit (and thus transform) after this index, possibly saving a fair bit of time
        last_fittable_idx = -1
        for idx, transformer in enumerate(self._transformers):
            if isinstance(transformer, FittableDataTransformer):
                last_fittable_idx = idx

        for idx, transformer in enumerate(self._transformers):
            if idx <= last_fittable_idx and isinstance(transformer, FittableDataTransformer):
                transformer.fit(data)

            if idx < last_fittable_idx:
                data = transformer.transform(data)

    def fit_transform(self, data: TimeSeries) -> TimeSeries:
        """
        For each data transformer in pipeline first fit the data if transformer is fittable then transform data using
        fitted transformer. The transformed data is then passed to next transformer.

        Parameters
        ----------
        data
            TimeSeries to fit and transform on.

        Returns
        -------
        TimeSeries
            Transformed data.
        """
        for transformer in self._transformers:
            if isinstance(transformer, FittableDataTransformer):
                transformer.fit(data)

            data = transformer.transform(data)
        return data

    def transform(self, data: TimeSeries) -> TimeSeries:
        """
        For each data transformer in pipeline transform data. Then transformed data is passed to next transformer.

        Parameters
        ----------
        data
            TimeSeries to be transformed.

        Returns
        -------
        TimeSeries
            Transformed data.
        """
        for transformer in self._transformers:
            data = transformer.transform(data)
        return data

    def inverse_transform(self, data: TimeSeries) -> TimeSeries:
        """
        For each data transformer in pipeline inverse_transform data. Then inverse transformed data is passed to next
        transformer. Transformers are traversed in reverse order. Raises value error if not all of the transformers are
        invertible.

        Parameters
        ----------
        data
            TimeSeries to be inverse transformed.

        Returns
        -------
        TimeSeries
            Inverse transformed data.
        """
        raise_if_not(self._invertible, "Not all transformers in the pipeline can perform inverse_transform", logger)

        for transformer in reversed(self._transformers):
            data = transformer.inverse_transform(data)
        return data

    def invertible(self) -> bool:
        """
        Returns whether the pipeline is invertible or not.
        A pipeline is invertible if all transformers in the pipeline are themselves invertible.

        Returns
        -------
        bool
            True if the pipeline is invertible, False otherwise
        """
        return self._invertible

    def __getitem__(self, key: Union[int, slice]) -> 'Pipeline':
        """
        Gets subset of Pipeline based either on index or slice with indexes.
        Resulting pipeline will deep copy transformers of the original pipeline.

        Parameters
        ----------
        key
            Either int or slice indicating the subset of data transformers to keep.

        Returns
        -------
        Pipeline
            Subset of pipeline determined by key.
        """
        raise_if_not(isinstance(key, int) or isinstance(key, slice), "key must be either an int or a slice", logger)

        if isinstance(key, int):
            transformers = [self._transformers[key]]
        else:
            transformers = self._transformers[key]
        return Pipeline(transformers, copy=True)

    def __iter__(self) -> Iterator[BaseDataTransformer[TimeSeries]]:
        """
        Returns
        -------
        Iterator
            Iterator on sequence of data transformers
        """
        return self._transformers.__iter__()

    def __len__(self):
        return len(self._transformers)

    def __copy__(self, deep: bool = True):
        return Pipeline(self._transformers, copy=deep)

    def __deepcopy__(self, memo=None):
        return self.__copy__(deep=True)

    def __str__(self):
        string = "Pipeline: "
        arrow = " -> "
        for transformer in self._transformers:
            string += str(transformer) + arrow

        return string[: -len(arrow)]

    def __repr__(self):
        return self.__str__()



import numpy as np
import pandas as pd

from .base_test_class import DartsBaseTestClass
from ..utils import timeseries_generation as tg
from ..metrics import r2_score
from ..models import StandardRegressionModel


def train_test_split(features, target, split_ts):
    """
    Splits all provided TimeSeries instances into train and test sets according to the provided timestamp.

    :param features: Feature TimeSeries instances to be split.
    :param target: Target TimeSeries instance to be split.
    :return: 4-tuple of the form (train_features, train_target, test_features, test_target)
    """

    # split features
    train_features = []
    test_features = []
    for feature in features:
        train_feature, test_feature = feature.split_after(split_ts)
        train_features.append(train_feature)
        test_features.append(test_feature)

    # split target
    train_target, test_target = target.split_after(split_ts)

    return (train_features, train_target, test_features, test_target)


def test_models_accuracy(test_case, models, features, target, min_r2):
    # for every model, test whether it predicts the target with a minimum r2 score of `min_r2`
    train_f, train_t, test_f, test_t = train_test_split(features, target, pd.Timestamp('20010101'))

    for model in models:
        model.fit(train_f, train_t)
        prediction = model.predict(test_f)
        current_r2 = r2_score(prediction, test_t)
        test_case.assertTrue(current_r2 >= min_r2, "{} model was not able to denoise data."
                             "A r2 score of {} was recorded.".format(str(model), current_r2))


class RegressionModelsTestCase(DartsBaseTestClass):

    np.random.seed(1)

    # number of data points used for training
    regression_window = 5

    # dummy feature and target TimeSeries instances
    ts_periodic = tg.sine_timeseries(length=500)
    ts_gaussian = tg.gaussian_timeseries(length=500)
    ts_random_walk = tg.random_walk_timeseries(length=500)
    ts_sum = ts_periodic + ts_gaussian
    ts_random_multi = ts_gaussian.stack(ts_random_walk)
    ts_sum_2 = ts_sum + ts_random_walk
    ts_sum_multi = ts_sum.stack(ts_sum_2)

    # default regression models
    models = [
        StandardRegressionModel(regression_window)
    ]

    def test_models_runnability(self):
        for model in self.models:
            # training and predicting on same features, since only runnability is tested
            model.fit([self.ts_periodic, self.ts_gaussian], self.ts_sum)
            prediction = model.predict([self.ts_periodic, self.ts_gaussian])
            self.assertTrue(len(prediction) == len(self.ts_periodic))

    def test_models_denoising(self):
        # for every model, test whether it correctly denoises ts_sum using ts_gaussian and ts_sum as inputs
        test_models_accuracy(self, self.models, [self.ts_gaussian, self.ts_sum], self.ts_periodic, 1.0)

    def test_models_denoising_multi_input(self):
        # for every model, test whether it correctly denoises ts_sum_2 using ts_random_multi and ts_sum_2 as inputs
        test_models_accuracy(self, self.models, [self.ts_random_multi, self.ts_sum_2], self.ts_periodic, 1.0)

    def test_models_denoising_multi_target(self):
        # for every model, test whether it correctly denoises ts_sum_multi using ts_random_multi and ts_sum_2 as inputs
        test_models_accuracy(self, self.models, [self.ts_random_multi, self.ts_sum_2], self.ts_sum_multi, 1.0)

    def test_wrong_dimensionality(self):
        train_f, train_t, _, _ = train_test_split([self.ts_periodic, self.ts_sum_multi],
                                                  self.ts_sum, pd.Timestamp('20010101'))
        self.models[0].fit(train_f, train_t)
        _, _, test_f, _ = train_test_split([self.ts_sum_multi, self.ts_periodic],
                                           self.ts_sum, pd.Timestamp('20010101'))

        with self.assertRaises(ValueError):
            self.models[0].predict(test_f)

    def test_displays_warning(self):
        series = tg.constant_timeseries(value=0, length=10)
        model = StandardRegressionModel(train_n_points=20)

        with self.assertWarns(UserWarning):
            model.fit([series], series)



"""
Missing Values Filler
---------------------
"""
from typing import Union

from darts.dataprocessing.transformers import BaseDataTransformer
from darts.utils.missing_values import fill_missing_values

from darts.timeseries import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not

logger = get_logger(__name__)


class MissingValuesFiller(BaseDataTransformer[TimeSeries]):
    def __init__(self,
                 fill: Union[str, float] = 'auto',
                 name: str = "MissingValuesFiller"):
        """
        Data transformer to fill missing values from time series

        Parameters
        ----------
        fill
            The value used to replace the missing values.
            If set to 'auto', will auto-fill missing values using the `pandas.Dataframe.interpolate()` method.
        name
            A specific name for the transformer
        """
        raise_if_not(isinstance(fill, str) or isinstance(fill, float),
                     "`fill` should either be a string or a float",
                     logger)
        raise_if(isinstance(fill, str) and fill != 'auto',
                 "invalid string for `fill`: can only be set to 'auto'",
                 logger)

        super().__init__(name)
        self._fill = fill

    def transform(self, data: TimeSeries, **interpolate_kwargs) -> TimeSeries:
        super().transform(data)
        return fill_missing_values(data, self._fill, **interpolate_kwargs)



import unittest
import pandas as pd

from darts.dataprocessing.transformers.mappers import Mapper, InvertibleMapper
from darts.utils.timeseries_generation import constant_timeseries, linear_timeseries


class MappersTestCase(unittest.TestCase):

    @staticmethod
    def func(x):
        return x + 10

    @staticmethod
    def inverse_func(x):
        return x - 10

    @staticmethod
    def ts_func(ts, x):
        return x - ts.month

    @staticmethod
    def inverse_ts_func(ts, x):
        return x + ts.month

    plus_ten = Mapper(func.__func__)
    plus_ten_invertible = InvertibleMapper(func.__func__, inverse_func.__func__)

    subtract_month = Mapper(ts_func.__func__)
    subtract_month_invertible = InvertibleMapper(ts_func.__func__, inverse_ts_func.__func__)

    lin_series = linear_timeseries(start_value=1, length=12, freq='MS', start_ts=pd.Timestamp('2000-01-01'), end_value=12)  # noqa: E501
    zeroes = constant_timeseries(value=0.0, length=12, freq='MS', start_ts=pd.Timestamp('2000-01-01'))
    tens = constant_timeseries(value=10.0, length=12, freq='MS', start_ts=pd.Timestamp('2000-01-01'))

    def test_mapper(self):
        transformed = self.plus_ten.transform(self.zeroes)
        self.assertEqual(transformed, self.tens)

    def test_invertible_mapper(self):
        transformed = self.plus_ten_invertible.transform(self.lin_series)
        back = self.plus_ten_invertible.inverse_transform(transformed)
        self.assertEqual(back, self.lin_series)

    def test_mapper_with_timestamp(self):
        transformed = self.subtract_month.transform(self.lin_series)
        self.assertEqual(transformed, self.zeroes)

    def test_invertible_mapper_with_timestamp(self):
        transformed = self.subtract_month_invertible.transform(self.lin_series)
        back = self.subtract_month_invertible.inverse_transform(transformed)
        self.assertEqual(back, self.lin_series)



"""
Scaler
------
"""
from darts.dataprocessing.transformers import FittableDataTransformer, InvertibleDataTransformer

from darts.timeseries import TimeSeries
from darts.logging import get_logger, raise_log
from sklearn.preprocessing import MinMaxScaler

logger = get_logger(__name__)


class Scaler(FittableDataTransformer[TimeSeries], InvertibleDataTransformer[TimeSeries]):
    def __init__(self, scaler=MinMaxScaler(feature_range=(0, 1)), name="Scaler"):
        """
        Generic wrapper class for using scalers that implement `fit()`, `transform()` and
        `inverse_transform()` methods (typically from scikit-learn) on `TimeSeries`.

        Parameters
        ----------
        scaler
            The scaler to transform the data.
            It must provide the `fit()`, `transform()` and `inverse_transform()` methods.
            Default: `sklearn.preprocessing.MinMaxScaler(feature_range=(0, 1))`; this
            will scale all the values of a time series between 0 and 1.
        name
            A specific name for the scaler
        """
        super().__init__(name)

        if (not callable(getattr(scaler, "fit", None)) or not callable(getattr(scaler, "transform", None))
                or not callable(getattr(scaler, "inverse_transform", None))): # noqa W503
            raise_log(ValueError(
                      'The provided transformer object must have fit(), transform() and inverse_transform() methods'),
                      logger)

        self.transformer = scaler
        self.train_series = None

    def fit(self, series: TimeSeries) -> 'Scaler':
        """
        Fits this scaler to the provided time series data

        Parameters
        ----------
        series
            The time series to fit the scaler on

        Returns
        -------
            Fitted scaler (self)
        """
        super().fit(series)
        self.transformer.fit(series.values().reshape((-1, series.width)))
        self.train_series = series
        return self

    def transform(self, series: TimeSeries, *args, **kwargs) -> TimeSeries:
        """
        Returns a new time series, transformed with this (fitted) scaler.
        This does not handle series with confidence intervals - the intervals are discarded.

        Parameters
        ----------
        series
            The time series to transform

        Returns
        -------
        TimeSeries
            A new time series, transformed with this (fitted) scaler.
        """
        super().transform(series, *args, **kwargs)
        return TimeSeries.from_times_and_values(series.time_index(),
                                                self.transformer.transform(series.values().
                                                                           reshape((-1, series.width))),
                                                series.freq())

    def inverse_transform(self, series: TimeSeries, *args, **kwargs) -> TimeSeries:
        """
        Performs the inverse transformation on a time series

        Parameters
        ----------
        series
            The time series to inverse transform

        Returns
        -------
        TimeSeries
            The inverse transform
        """
        super().inverse_transform(series, *args, **kwargs)
        return TimeSeries.from_times_and_values(series.time_index(),
                                                self.transformer.inverse_transform(series.values().
                                                                                   reshape((-1, series.width))),
                                                series.freq())



"""
Regression Model Base Class
---------------------------

A regression model predicts values for a time series :math:`Y_t` as a function
of :math:`N` "features" time series :math:`X^i_t`:

.. math:: Y_t = f(X^1_t, ..., X^N_t),

where :math:`t` denotes the time step. Here, the function :math:`f()` is not necessarily linear.
"""

import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import List, Iterable, Union, Any, Callable

from ..timeseries import TimeSeries
from ..logging import raise_if_not, get_logger, raise_log
from ..utils import (
    _build_tqdm_iterator,
    _with_sanity_checks,
    _historical_forecasts_general_checks
)
from .. import metrics

logger = get_logger(__name__)


# TODO: Extend this to a "DynamicRegressiveModel" class, which acts on List[List[TimeSeries]].
# TODO: The first List[] would contain time-sliding lists of time series, letting the model
# TODO: be able to learn how to change weights over time. When len() of outer List[] is 0 it's a particular case
class RegressionModel(ABC):
    @abstractmethod
    def __init__(self):
        """ Regression Model.

            This is the base class for all regression models.
        """

        # Stores training date information:
        self.train_features: List[TimeSeries] = None
        self.train_target: TimeSeries = None

        # state
        self._fit_called = False

    @abstractmethod
    def fit(self, train_features: List[TimeSeries], train_target: TimeSeries) -> None:
        """ Fits/trains the model using the provided list of features time series and the target time series.

        Parameters
        ----------
        train_features
            A list of features time series, all of the same length as the target series
        train_target
            A target time series, of the same length as the features series
        """

        raise_if_not(len(train_features) > 0, 'Need at least one feature series', logger)
        raise_if_not(all([s.has_same_time_as(train_target) for s in train_features]),
                     'All provided time series must have the same time index', logger)
        self.train_features = train_features
        self.train_target = train_target
        self._fit_called = True

    @abstractmethod
    def predict(self, features: List[TimeSeries]) -> TimeSeries:
        """ Predicts values of the target time series, given a list of features time series

        Parameters
        ----------
        features
            The list of features time series, of the same length

        Returns
        -------
        TimeSeries
            A series containing the predicted targets, of the same length as the features series
        """

        if (not self._fit_called):
            raise_log(Exception('fit() must be called before predict()'), logger)

        length_ok = len(features) == len(self.train_features)
        dimensions_ok = all(features[i].width == self.train_features[i].width for i in range(len(features)))
        raise_if_not(length_ok and dimensions_ok,
                     'The number and dimensionalities of all given features must correspond to those used for'
                     ' training.', logger)

    def _historical_forecasts_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Sanity checks for the historical_forecasts function

        Parameters
        ----------
        args
            The args parameter(s) provided to the historical_forecasts function.
        kwargs
            The kwargs paramter(s) provided to the historical_forecasts function.

        Raises
        ------
        ValueError
            when a check on the parameter does not pass.
        """

        # parse args
        feature_series = args[0]
        target_series = args[1]

        raise_if_not(all([s.has_same_time_as(target_series) for s in feature_series]), 'All provided time series must '
                     'have the same time index', logger)

        _historical_forecasts_general_checks(target_series, kwargs)

    @_with_sanity_checks("_historical_forecasts_sanity_checks")
    def historical_forecasts(self,
                             feature_series: Iterable[TimeSeries],
                             target_series: TimeSeries,
                             start: Union[pd.Timestamp, float, int] = 0.5,
                             forecast_horizon: int = 1,
                             stride: int = 1,
                             overlap_end: bool = False,
                             last_points_only: bool = True,
                             verbose: bool = False) -> Union[List[TimeSeries], TimeSeries]:
        """ Computes the historical forecasts the model would have produced with an expanding training window
        and (by default) returns a time series created from the last point of each of these individual forecasts

        To this end, it repeatedly builds a training set composed of both features and targets,
        from `feature_series` and `target_series`, respectively.
        It trains the current model on the training set, emits a forecast of length equal to forecast_horizon,
        and then moves the end of the training set forward by `stride` time steps.

        By default, this method will return a single time series made up of the last point of each
        historical forecast. This time series will thus have a frequency of training_series.freq() * stride
        If `last_points_only` is set to False, it will instead return a list of the historical forecasts.

        This always re-trains the models on the entire available history,
        corresponding an expanding window strategy.

        Parameters
        ----------
        feature_series
            An iterable of time series representing the features for the regression model (independent variables)
        target_series
            The univariate target time series for the regression model (dependent variable)
        start
            The first prediction time, at which a prediction is computed for a future time
        forecast_horizon
            The forecast horizon for the point predictions
        stride
            The number of time steps between two consecutive predictions.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not
        last_points_only
            Whether to retain only the last point of each historical forecast.
            If set to True, the method returns a single `TimeSeries` of the point forecasts.
            Otherwise returns a list of historical `TimeSeries` forecasts.
        verbose
            Whether to print progress

        Returns
        -------
        TimeSeries or List[TimeSeries]
            By default, a single TimeSeries instance created from the last point of each individual forecast.
            If `last_points_only` is set to False, a list of the historical forecasts
        """
        start = target_series.get_timestamp_at_point(start)

        # build the prediction times in advance (to be able to use tqdm)
        if not overlap_end:
            last_valid_pred_time = target_series.time_index()[-1 - forecast_horizon]
        else:
            last_valid_pred_time = target_series.time_index()[-2]

        pred_times = [start]
        while pred_times[-1] < last_valid_pred_time:
            # compute the next prediction time and add it to pred times
            pred_times.append(pred_times[-1] + target_series.freq() * stride)

        # the last prediction time computed might have overshot last_valid_pred_time
        if pred_times[-1] > last_valid_pred_time:
            pred_times.pop(-1)

        iterator = _build_tqdm_iterator(pred_times, verbose)

        # Either store the whole forecasts or only the last points of each forecast, depending on last_points_only
        forecasts = []

        last_points_times = []
        last_points_values = []

        for pred_time in iterator:
            # build train/val series
            train_features = [s.drop_after(pred_time) for s in feature_series]
            train_target = target_series.drop_after(pred_time)
            val_features = [s.slice_n_points_after(pred_time, forecast_horizon) for s in feature_series]

            self.fit(train_features, train_target)
            forecast = self.predict(val_features)

            if last_points_only:
                last_points_values.append(forecast.values()[-1])
                last_points_times.append(forecast.end_time())
            else:
                forecasts.append(forecast)

        if last_points_only:
            return TimeSeries.from_times_and_values(pd.DatetimeIndex(last_points_times),
                                                    np.array(last_points_values),
                                                    freq=target_series.freq() * stride)

        return forecasts

    def backtest(self,
                 feature_series: Iterable[TimeSeries],
                 target_series: TimeSeries,
                 start: Union[pd.Timestamp, float, int] = 0.5,
                 forecast_horizon: int = 1,
                 stride: int = 1,
                 overlap_end: bool = False,
                 last_points_only: bool = False,
                 metric: Callable[[TimeSeries, TimeSeries], float] = metrics.mape,
                 reduction: Union[Callable[[np.ndarray], float], None] = np.mean,
                 verbose: bool = False) -> Union[float, List[float]]:
        """Computes an error score between the historical forecasts the model would have produced
        with an expanding training window over `series` and the actual series.

        To this end, it repeatedly builds a training set composed of both features and targets,
        from `feature_series` and `target_series`, respectively.
        It trains the current model on the training set, emits a forecast of length equal to forecast_horizon,
        and then moves the end of the training set forward by `stride` time steps.

        By default, this method will use each historical forecast (whole) to compute error scores.
        If `last_points_only` is set to True, it will use only the last point of each historical forecast.

        This always re-trains the models on the entire available history,
        corresponding an expanding window strategy.

        Parameters
        ----------
        feature_series
            An iterable of time series representing the features for the regression model (independent variables)
        target_series
            The univariate target time series for the regression model (dependent variable)
        start
            The first prediction time, at which a prediction is computed for a future time
        forecast_horizon
            The forecast horizon for the point predictions
        stride
            The number of time steps between two consecutive predictions.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not
        last_points_only
            Whether to keep the whole historical forecasts or only the last point of each forecast
        metric
            A function that takes two TimeSeries instances as inputs and returns a float error value.
        reduction
            A function used to combine the individual error scores obtained when `last_points_only` is set to False.
            If explicitely set to `None`, the method will return a list of the individual error scores instead.
            Set to np.mean by default.
        verbose
            Whether to print progress

        Returns
        -------
        float or List[float]
            The error score, or the list of individual error scores if `reduction` is `None`
        """
        forecasts = self.historical_forecasts(feature_series,
                                              target_series,
                                              start,
                                              forecast_horizon,
                                              stride,
                                              overlap_end,
                                              last_points_only,
                                              verbose)

        if last_points_only:
            return metric(target_series, forecasts)

        errors = []
        for forecast in forecasts:
            errors.append(metric(target_series, forecast))

        if reduction is None:
            return errors

        return reduction(errors)

    def residuals(self) -> TimeSeries:
        """ Computes the time series of residuals of this model on the training time series

        The residuals are computed as

        .. math:: z_t := y_t - \\hat{y}_t,

        where :math:`y_t` is the actual target time series over the training set,
        and :math:`\\hat{y}_t` is the time series of predicted targets, over the training set.

        Returns
        -------
        TimeSeries
            The time series containing the residuals
        """

        if (not self._fit_called):
            raise_log(Exception('fit() must be called before predict()'), logger)

        train_pred = self.predict(self.train_features)
        return self.train_target - train_pred



import unittest
import logging

from darts.dataprocessing.transformers import BaseDataTransformer


class BaseDataTransformerTestCase(unittest.TestCase):
    __test__ = True

    @classmethod
    def setUpClass(cls):
        logging.disable(logging.CRITICAL)

    class DataTransformerMock(BaseDataTransformer[str]):
        def __init__(self):
            super().__init__(name="DataTransformerMock")
            self.transform_called = False

        def transform(self, data: str, *args, **kwargs) -> str:
            self.transform_called = True
            return data + "transformed"

    def test_input_transformed(self):
        # given
        test_input = "test"
        mock = self.DataTransformerMock()

        # when
        transformed = mock.transform(test_input)

        expected = "testtransformed"
        self.assertEqual(transformed, expected)



"""
Box-Cox Transformer
-------------------
"""

from typing import Optional, Union, Sequence
from scipy.stats import boxcox_normmax, boxcox
from scipy.special import inv_boxcox

from darts.timeseries import TimeSeries
from darts.dataprocessing.transformers import FittableDataTransformer, InvertibleDataTransformer
from darts.logging import get_logger, raise_if

logger = get_logger(__name__)


class BoxCox(FittableDataTransformer[TimeSeries], InvertibleDataTransformer[TimeSeries]):

    def __init__(self,
                 name: str = "BoxCox"):
        """
        Box-Cox data transformer.
        See https://otexts.com/fpp2/transformations.html#mathematical-transformations for more information

        Parameters
        ----------
        name
            A specific name for the transformer
        """
        super().__init__(name)
        self._lmbda = None

    def fit(self,
            data: TimeSeries,
            lmbda: Optional[Union[float, Sequence[float]]] = None,
            optim_method='mle') -> 'BoxCox':
        """
        Sets the `lmbda` parameter value.

        Parameters
        ----------
        data
            The time series to fit on
        lmbda
            If None given, will automatically find an optimal value of lmbda (for each dimension
            of the time series) using `scipy.stats.boxcox_normmax` with `method=optim_method`
            If a single float is given, the same lmbda value will be used for all dimensions of the series.
            Also allows to specify a different lmbda value for each dimension of the time series by passing
            a sequence of values.
        optim_method
            Specifies which method to use to find an optimal value for the lmbda parameter.
            Either 'mle' or 'pearsonr'.

        Returns
        -------
            Fitted transformer (self)
        """
        super().fit(data)

        raise_if(not isinstance(optim_method, str) or optim_method not in ['mle', 'pearsonr'],
                 "optim_method parameter must be either 'mle' or 'pearsonr'",
                 logger)

        if lmbda is None:
            # Compute optimal lmbda for each dimension of the time series
            lmbda = data._df.apply(boxcox_normmax, method=optim_method)
        elif isinstance(lmbda, Sequence):
            raise_if(len(lmbda) != data.width,
                     "lmbda should have one value per dimension (ie. column or variable) of the time series",
                     logger)
        else:
            # Replicate lmbda to match dimensions of the time series
            lmbda = [lmbda] * data.width

        self._lmbda = lmbda

        return self

    def transform(self, data: TimeSeries, *args, **kwargs) -> TimeSeries:
        super().transform(data, *args, **kwargs)

        def _boxcox_wrapper(col):
            idx = data._df.columns.get_loc(col.name)  # get index from col name
            return boxcox(col, self._lmbda[idx])

        return TimeSeries.from_dataframe(data._df.apply(_boxcox_wrapper))

    def inverse_transform(self, data: TimeSeries, *args, **kwargs) -> TimeSeries:
        super().inverse_transform(data, *args, *kwargs)

        def _inv_boxcox_wrapper(col):
            idx = data._df.columns.get_loc(col.name)  # get index from col name
            return inv_boxcox(col, self._lmbda[idx])

        return TimeSeries.from_dataframe(data._df.apply(_inv_boxcox_wrapper))



"""
N-BEATS
-------
"""

from typing import NewType, Optional, Union, List
from enum import Enum
import numpy as np
from numpy.random import RandomState
import torch
import torch.nn as nn

from ..logging import get_logger, raise_log, raise_if_not
from ..utils.torch import random_method
from .torch_forecasting_model import TorchForecastingModel

logger = get_logger(__name__)


class _GType(Enum):
    GENERIC = 1
    TREND = 2
    SEASONALITY = 3


GTypes = NewType('GTypes', _GType)


class _TrendGenerator(nn.Module):

    def __init__(self,
                 expansion_coefficient_dim,
                 output_dim):
        super(_TrendGenerator, self).__init__()
        self.T = torch.stack([(torch.arange(output_dim) / output_dim)**i for i in range(expansion_coefficient_dim)], 1)

    def forward(self, x):
        return torch.matmul(x, self.T.float().T)


class _SeasonalityGenerator(nn.Module):

    def __init__(self,
                 output_dim):
        super(_SeasonalityGenerator, self).__init__()
        half_minus_one = int(output_dim / 2 - 1)
        cos_vectors = [torch.cos(torch.arange(output_dim) * 2 * np.pi * i) for i in range(1, half_minus_one + 1)]
        sin_vectors = [torch.sin(torch.arange(output_dim) * 2 * np.pi * i) for i in range(1, half_minus_one + 1)]
        self.S = torch.stack([torch.ones(output_dim)] + cos_vectors + sin_vectors, 1)

    def forward(self, x):
        return torch.matmul(x, self.S.float().T)


class _Block(nn.Module):

    def __init__(self,
                 num_layers: int,
                 layer_width: int,
                 expansion_coefficient_dim: int,
                 input_chunk_length: int,
                 target_length: int,
                 g_type: GTypes):
        """ PyTorch module implementing the basic building block of the N-BEATS architecture.

        Parameters
        ----------
        num_layers
            The number of fully connected layers preceding the final forking layers.
        layer_width
            The number of neurons that make up each fully connected layer.
        expansion_coefficient_dim
            The dimensionality of the waveform generator parameters, also known as expansion coefficients.
            Only used if `generic_architecture` is set to `True`.
        input_chunk_length
            The length of the input sequence fed to the model.
        target_length
            The length of the forecast of the model.
        g_type
            The type of function that is implemented by the waveform generator.

        Inputs
        ------
        x of shape `(batch_size, input_chunk_length)`
            Tensor containing the input sequence.

        Outputs
        -------
        x_hat of shape `(batch_size, input_chunk_length)`
            Tensor containing the 'backcast' of the block, which represents an approximation of `x`
            given the constraints of the functional space determined by `g`.
        y_hat of shape `(batch_size, output_chunk_length)`
            Tensor containing the forward forecast of the block.

        """
        super(_Block, self).__init__()

        self.num_layers = num_layers
        self.layer_width = layer_width
        self.g_type = g_type
        self.relu = nn.ReLU()

        # fully connected stack before fork
        self.linear_layer_stack_list = [nn.Linear(input_chunk_length, layer_width)]
        self.linear_layer_stack_list += [nn.Linear(layer_width, layer_width) for _ in range(num_layers - 1)]
        self.fc_stack = nn.ModuleList(self.linear_layer_stack_list)

        # fully connected layer producing forecast/backcast expansion coeffcients (waveform generator parameters)
        if g_type == _GType.SEASONALITY:
            self.backcast_linear_layer = nn.Linear(layer_width, 2 * int(input_chunk_length / 2 - 1) + 1)
            self.forecast_linear_layer = nn.Linear(layer_width, 2 * int(target_length / 2 - 1) + 1)
        else:
            self.backcast_linear_layer = nn.Linear(layer_width, expansion_coefficient_dim)
            self.forecast_linear_layer = nn.Linear(layer_width, expansion_coefficient_dim)

        # waveform generator functions
        if g_type == _GType.GENERIC:
            self.backcast_g = nn.Linear(expansion_coefficient_dim, input_chunk_length)
            self.forecast_g = nn.Linear(expansion_coefficient_dim, target_length)
        elif g_type == _GType.TREND:
            self.backcast_g = _TrendGenerator(expansion_coefficient_dim, input_chunk_length)
            self.forecast_g = _TrendGenerator(expansion_coefficient_dim, target_length)
        elif g_type == _GType.SEASONALITY:
            self.backcast_g = _SeasonalityGenerator(input_chunk_length)
            self.forecast_g = _SeasonalityGenerator(target_length)
        else:
            raise_log(ValueError("g_type not supported"), logger)

    def forward(self, x):
        # fully connected layer stack
        for layer in self.linear_layer_stack_list:
            x = self.relu(layer(x))

        # forked linear layers producing waveform generator parameters
        theta_backcast = self.backcast_linear_layer(x)
        theta_forecast = self.forecast_linear_layer(x)

        # waveform generator applications
        x_hat = self.backcast_g(theta_backcast)
        y_hat = self.forecast_g(theta_forecast)

        return x_hat, y_hat


class _Stack(nn.Module):

    def __init__(self,
                 num_blocks: int,
                 num_layers: int,
                 layer_width: int,
                 expansion_coefficient_dim: int,
                 input_chunk_length: int,
                 target_length: int,
                 g_type: GTypes,
                 ):
        """ PyTorch module implementing one stack of the N-BEATS architecture that comprises multiple basic blocks.

        Parameters
        ----------
        num_blocks
            The number of blocks making up this stack.
        num_layers
            The number of fully connected layers preceding the final forking layers in each block.
        layer_width
            The number of neurons that make up each fully connected layer in each block.
        expansion_coefficient_dim
            The dimensionality of the waveform generator parameters, also known as expansion coefficients.
            Only used if `generic_architecture` is set to `True`.
        input_chunk_length
            The length of the input sequence fed to the model.
        target_length
            The length of the forecast of the model.
        g_type
            The function that is implemented by the waveform generators in each block.

        Inputs
        ------
        stack_input of shape `(batch_size, input_chunk_length)`
            Tensor containing the input sequence.

        Outputs
        -------
        stack_residual of shape `(batch_size, input_chunk_length)`
            Tensor containing the 'backcast' of the block, which represents an approximation of `x`
            given the constraints of the functional space determined by `g`.
        stack_forecast of shape `(batch_size, output_chunk_length)`
            Tensor containing the forward forecast of the stack.

        """
        super(_Stack, self).__init__()

        self.input_chunk_length = input_chunk_length
        self.target_length = target_length

        if g_type == _GType.GENERIC:
            self.blocks_list = [
                _Block(num_layers, layer_width, expansion_coefficient_dim, input_chunk_length, target_length, g_type)
                for _ in range(num_blocks)
            ]
        else:
            # same block instance is used for weight sharing
            interpretable_block = _Block(num_layers, layer_width, expansion_coefficient_dim,
                                         input_chunk_length, target_length, g_type)
            self.blocks_list = [interpretable_block] * num_blocks

        self.blocks = nn.ModuleList(self.blocks_list)

    def forward(self, x):
        stack_forecast = torch.zeros(x.shape[0], self.target_length, device=x.device)
        for block in self.blocks_list:
            # pass input through block
            x_hat, y_hat = block(x)

            # add block forecast to stack forecast
            stack_forecast = stack_forecast + y_hat

            # subtract backcast from input to produce residual
            x = x - x_hat

        stack_residual = x

        return stack_residual, stack_forecast


class _NBEATSModule(nn.Module):

    def __init__(self,
                 input_chunk_length: int,
                 output_chunk_length: int,
                 generic_architecture: bool,
                 num_stacks: int,
                 num_blocks: int,
                 num_layers: int,
                 layer_widths: List[int],
                 expansion_coefficient_dim: int,
                 trend_polynomial_degree: int
                 ):
        """ PyTorch module implementing the N-BEATS architecture.

        Parameters
        ----------
        input_chunk_length
            The length of the input sequence fed to the model.
        output_chunk_length
            The length of the forecast of the model.
        generic_architecture
            Boolean value indicating whether the generic architecture of N-BEATS is used.
            If not, the interpretable architecture outlined in the paper (consisting of one trend
            and one seasonality stack with appropriate waveform generator functions).
        num_stacks
            The number of stacks that make up the whole model. Only used if `generic_architecture` is set to `True`.
        num_blocks
            The number of blocks making up every stack.
        num_layers
            The number of fully connected layers preceding the final forking layers in each block of every stack.
            Only used if `generic_architecture` is set to `True`.
        layer_widths
            Determines the number of neurons that make up each fully connected layer in each block of every stack.
            If a list is passed, it must have a length equal to `num_stacks` and every entry in that list corresponds
            to the layer width of the corresponding stack. If an integer is passed, every stack will have blocks
            with FC layers of the same width.
        expansion_coefficient_dim
            The dimensionality of the waveform generator parameters, also known as expansion coefficients.
            Only used if `generic_architecture` is set to `True`.
        trend_polynomial_degree
            The degree of the polynomial used as waveform generator in trend stacks. Only used if
            `generic_architecture` is set to `False`.

        Inputs
        ------
        x of shape `(batch_size, input_chunk_length)`
            Tensor containing the input sequence.

        Outputs
        -------
        y of shape `(batch_size, output_chunk_length)`
            Tensor containing the output of the NBEATS module.

        """
        super(_NBEATSModule, self).__init__()

        self.input_chunk_length = input_chunk_length
        self.target_length = output_chunk_length

        if generic_architecture:
            self.stacks_list = [
                _Stack(num_blocks, num_layers, layer_widths[i], expansion_coefficient_dim,
                       input_chunk_length, output_chunk_length, _GType.GENERIC) for i in range(num_stacks)
            ]
        else:
            num_stacks = 2
            trend_stack = _Stack(num_blocks, num_layers, layer_widths[0], trend_polynomial_degree + 1,
                                 input_chunk_length, output_chunk_length, _GType.TREND)
            seasonality_stack = _Stack(num_blocks, num_layers, layer_widths[1], -1,
                                       input_chunk_length, output_chunk_length, _GType.SEASONALITY)
            self.stacks_list = [trend_stack, seasonality_stack]

        self.stacks = nn.ModuleList(self.stacks_list)

    def forward(self, x):

        # squeeze last dimension (because model is univariate)
        x = x.squeeze(dim=2)

        y = torch.zeros(x.shape[0], self.target_length, device=x.device)
        for stack in self.stacks_list:
            # compute stack output
            stack_residual, stack_forecast = stack(x)

            # add stack forecast to final output
            y = y + stack_forecast

            # set current stack residual as input for next stack
            x = stack_residual

        # unsqueeze last dimension
        y = y.unsqueeze(dim=2)

        return y


class NBEATSModel(TorchForecastingModel):
    @random_method
    def __init__(self,
                 input_chunk_length: int,
                 output_chunk_length: int,
                 generic_architecture: bool = True,
                 num_stacks: int = 30,
                 num_blocks: int = 1,
                 num_layers: int = 4,
                 layer_widths: Union[int, List[int]] = 256,
                 expansion_coefficient_dim: int = 5,
                 trend_polynomial_degree: int = 2,
                 **kwargs):
        """ Neural Basis Expansion Analysis Time Series Forecasting (N-BEATS).

        This is an implementation of the N-BEATS architecture, as outlined in this paper:
        https://openreview.net/forum?id=r1ecqn4YwB

        This model currently supports only univariate time series (and thus doesn't support using covariates).
        It can nevertheless be trained on several time series.

        Parameters
        ----------
        input_chunk_length
            The length of the input sequence fed to the model.
        output_chunk_length
            The length of the forecast of the model.
        generic_architecture
            Boolean value indicating whether the generic architecture of N-BEATS is used.
            If not, the interpretable architecture outlined in the paper (consisting of one trend
            and one seasonality stack with appropriate waveform generator functions).
        num_stacks
            The number of stacks that make up the whole model. Only used if `generic_architecture` is set to `True`.
            The interpretable architecture always uses two stacks - one for trend and one for seasonality.
        num_blocks
            The number of blocks making up every stack.
        num_layers
            The number of fully connected layers preceding the final forking layers in each block of every stack.
            Only used if `generic_architecture` is set to `True`.
        layer_widths
            Determines the number of neurons that make up each fully connected layer in each block of every stack.
            If a list is passed, it must have a length equal to `num_stacks` and every entry in that list corresponds
            to the layer width of the corresponding stack. If an integer is passed, every stack will have blocks
            with FC layers of the same width.
        expansion_coefficient_dim
            The dimensionality of the waveform generator parameters, also known as expansion coefficients.
            Only used if `generic_architecture` is set to `True`.
        trend_polynomial_degree
            The degree of the polynomial used as waveform generator in trend stacks. Only used if
            `generic_architecture` is set to `False`.
        random_state
            Control the randomness of the weights initialization. Check this
            `link <https://scikit-learn.org/stable/glossary.html#term-random-state>`_ for more details.

        """

        kwargs['input_chunk_length'] = input_chunk_length
        kwargs['output_chunk_length'] = output_chunk_length
        super().__init__(**kwargs)

        raise_if_not(isinstance(layer_widths, int) or len(layer_widths) == num_stacks,
                     "Please pass an integer or a list of integers with length `num_stacks`"
                     "as value for the `layer_widths` argument.", logger)

        self.input_chunk_length = input_chunk_length
        self.output_chunk_length = output_chunk_length
        self.generic_architecture = generic_architecture
        self.num_stacks = num_stacks
        self.num_blocks = num_blocks
        self.num_layers = num_layers
        self.layer_widths = layer_widths
        self.expansion_coefficient_dim = expansion_coefficient_dim
        self.trend_polynomial_degree = trend_polynomial_degree

        if not generic_architecture:
            self.num_stacks = 2

        if isinstance(layer_widths, int):
            self.layer_widths = [layer_widths] * num_stacks

    def _create_model(self, input_dim: int, output_dim: int) -> torch.nn.Module:
        raise_if_not(input_dim == 1 and output_dim == 1,
                     'The N-Beats model currently supports only univariate time series.'
                     'Currently: input_dim = {} and  output_dim = {}'.format(input_dim, output_dim))
        return _NBEATSModule(
            self.input_chunk_length,
            self.output_chunk_length,
            self.generic_architecture,
            self.num_stacks,
            self.num_blocks,
            self.num_layers,
            self.layer_widths,
            self.expansion_coefficient_dim,
            self.trend_polynomial_degree
        )



"""
Fittable Data Transformer
-------------------------
"""

from abc import abstractmethod
from typing import TypeVar

from darts.logging import get_logger, raise_if_not
from darts.dataprocessing.transformers import BaseDataTransformer

logger = get_logger(__name__)
T = TypeVar('T')


class FittableDataTransformer(BaseDataTransformer[T]):
    def __init__(self,
                 name: str = "FittableDataTransformer"):

        """
        Abstract class for data transformers implementing a fit method. All deriving classes must implement
        `fit()` and `transform()`.

        names
            The data transformer's name
        """
        super().__init__(name)
        self._fit_called = False

    @abstractmethod
    def fit(self, data: T) -> 'FittableDataTransformer[T]':
        """
        Fit the data transformer to data.
        Not implemented in base class and has to be implemented by deriving classes.

        Parameters
        ----------
        data
            Object on which data transformer will be fitted.

        Returns
        -------
        BaseDataTransformer[T]
            Fitted data transformer (typically would be self)
        """
        self._fit_called = True

    @abstractmethod
    def transform(self, data: T, *args, **kwargs) -> T:
        """
        Transform the data. Not implemented in base class.
        Will raise an error if called before a call to `fit()`

        Parameters
        ----------
        data
            Object which will be inverse transformed.
        args
            Additional positional arguments for the `transform()` method
        kwargs
            Additional keyword arguments for the `transform()` method

        Returns
        -------
        T
            Transformed data.
        """
        raise_if_not(self._fit_called, "fit() must have been called before transform()", logger)
        super().transform(data, args, kwargs)

    def fit_transform(self, data: T, *args, **kwargs) -> T:
        """
        Fit the transformer to data and then transform data.

        Parameters
        ----------
        data
            Object used to fit and transform.
        args
            Additional positional arguments for the `transform()` method
        kwargs
            Additional keyword arguments for the `transform()` method

        Returns
        -------
        T
            Transformed data.
        """
        return self.fit(data).transform(data, *args, **kwargs)



"""
Base Data Transformer
---------------------
"""
from abc import ABC, abstractmethod
from typing import TypeVar, Generic

T = TypeVar('T')


class BaseDataTransformer(Generic[T], ABC):
    def __init__(self,
                 name: str = "BaseDataTransformer"):
        """
        Abstract class for data transformers. All deriving classes have to implement only one function `transform`.
        Data transformers requiring to be fit first before calling `transform()` should derive
        from `FittableDataTransformer` instead.
        Data transformers which are invertible should derive from ´InvertibleDataTransformer´ instead.

        Parameters
        ----------
        names
            The data transformer's name
        """
        self._name = name

    @abstractmethod
    def transform(self, data: T, *args, **kwargs) -> T:
        """
        Transform the data.
        Not implemented in base class and has to be implemented by deriving classes.

        Parameters
        ----------
        data
            Object which will be transformed.
        args
            Additional positional arguments for the `transform` method
        kwargs
            Additional keyword arguments for the `transform` method

        Returns
        -------
        T
            Transformed data.
        """
        pass

    @property
    def name(self):
        """
        Returns
        -------
        str
            Name of data transformer.
        """
        return self._name

    def __str__(self):
        return self._name

    def __repr__(self):
        return self.__str__()



import unittest
import pandas as pd
from math import log

from darts.dataprocessing.transformers import BoxCox, Mapper
from darts.utils.timeseries_generation import sine_timeseries, linear_timeseries


class BoxCoxTestCase(unittest.TestCase):

    sine_series = sine_timeseries(length=50, value_y_offset=5, value_frequency=0.05)
    lin_series = linear_timeseries(start_value=1, end_value=10, length=50)
    multi_series = sine_series.stack(lin_series)

    def test_boxbox_lambda(self):
        boxcox = BoxCox()

        boxcox.fit(self.multi_series, 0.3)
        self.assertEqual(boxcox._lmbda, [0.3, 0.3])

        boxcox.fit(self.multi_series, [0.3, 0.4])
        self.assertEqual(boxcox._lmbda, [0.3, 0.4])

        with self.assertRaises(ValueError):
            boxcox.fit(self.multi_series, [0.2, 0.4, 0.5])

        boxcox.fit(self.multi_series, optim_method='mle')
        lmbda1 = boxcox._lmbda
        boxcox.fit(self.multi_series, optim_method='pearsonr')
        lmbda2 = boxcox._lmbda

        self.assertNotEqual(lmbda1.array, lmbda2.array)

    def test_boxcox_transform(self):
        log_mapper = Mapper(lambda x: log(x))
        boxcox = BoxCox()

        transformed1 = log_mapper.transform(self.sine_series)
        transformed2 = boxcox.fit(self.sine_series, lmbda=0).transform(self.sine_series)

        self.assertEqual(transformed1, transformed2)

    def test_boxcox_inverse(self):
        boxcox = BoxCox()
        transformed = boxcox.fit_transform(self.multi_series)
        back = boxcox.inverse_transform(transformed)
        pd.testing.assert_frame_equal(self.multi_series._df, back._df, check_exact=False)
